import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matching.mentor import Mentor
    from matching.mentee import Mentee


class Match:
    def __init__(self, mentor: "Mentor", mentee: "Mentee", weightings=None):
        self.weightings = (
            {"profession": 4, "grade": 3, "unmatched bonus": 0}
            if weightings is None
            else weightings
        )
        self.mentee = mentee
        self.mentor = mentor
        self._disallowed: bool = False
        self._score: int = 0
        self.calculate_match()

    @property
    def score(self):
        if self._disallowed:
            return 0
        else:
            return self._score

    @score.setter
    def score(self, new_value: int):
        self._score += new_value

    @property
    def disallowed(self):
        return self._disallowed

    @disallowed.setter
    def disallowed(self, new_value: bool):
        if self._disallowed is False and new_value is True:
            self._disallowed = new_value

    def calculate_match(self) -> None:
        scoring_methods = [
            self.check_not_already_matched,
            self.score_department,
            self.score_grade,
            self.score_profession,
            self.score_unmatched,
        ]
        while not self._disallowed and scoring_methods:
            scoring_method = scoring_methods.pop()
            scoring_method()

    def score_grade(self) -> None:
        grade_diff = self.mentor.grade - self.mentee.grade
        if not (2 >= grade_diff > 0):
            self._disallowed = True
        else:
            self._score += grade_diff * self.weightings["grade"]

    def score_profession(self) -> None:
        if self.mentee.profession == self.mentor.profession:
            self._score += self.weightings["profession"]

    def score_department(self) -> None:
        if self.mentee.department == self.mentor.department:
            self._disallowed = True

    def score_unmatched(self) -> None:
        if any(
            map(
                lambda participant: len(participant.connections) == 0,
                (self.mentee, self.mentor),
            )
        ):
            self._score += self.weightings.get("unmatched bonus")

    def mark_successful(self):
        if not self.disallowed:
            self.mentor.mentees.append(self.mentee)
            self.mentee.mentors.append(self.mentor)
        else:
            logging.debug("Skipping this match as disallowed")

    def check_not_already_matched(self):
        if self.mentee in self.mentor.mentees or self.mentor in self.mentee.mentors:
            self._disallowed = True



from typing import Dict, List

from matching.mentee import Mentee
from matching.mentor import Mentor
from matching.person import GRADES


class ParticipantFactory:
    @classmethod
    def create_from_dict(cls, data_as_dict):
        if "mentor" in data_as_dict:
            participant_type = Mentor
        elif "mentee" in data_as_dict:
            participant_type = Mentee
        else:
            raise TypeError
        participant_data = data_as_dict[participant_type.__name__.lower()]
        participant = participant_type()
        participant._grade = GRADES.index(participant_data["grade"])
        participant.department = participant_data["department"]
        participant.profession = participant_data["profession"]
        participant.email = participant_data["email"]
        participant.first_name = participant_data["first name"]
        participant.last_name = participant_data["last name"]
        participant.role = participant_data["role"]
        connections: List[Dict[str, str]] = data_as_dict.get("connections", [])
        participant._connections = [
            ParticipantFactory.create_from_dict(connection_data)
            for connection_data in connections
        ]
        return participant



import csv
import logging
import os
import sys
from functools import reduce
from pathlib import Path
from typing import Union, Type, List, Dict, Tuple, Generator, Optional

from munkres import Munkres, make_cost_matrix, Matrix  # type: ignore

from matching.match import Match
from matching.mentee import Mentee
from matching.mentor import Mentor


def generate_match_matrix(
    mentor_list: List[Mentor],
    mentee_list: List[Mentee],
    weightings: Optional[Dict[str, int]],
) -> List[List[Match]]:
    return [
        [Match(mentor, mentee, weightings) for mentee in mentee_list]
        for mentor in mentor_list
    ]


def process_form(path_to_form) -> Generator[Dict[str, str], None, None]:
    with open(path_to_form, "r") as data_form:
        file_reader = csv.DictReader(data_form)
        for row in file_reader:
            yield row


def create_participant_list_from_path(
    participant: Union[Type[Mentee], Type[Mentor]], path_to_data
):
    path_to_data = path_to_data / f"{participant.__name__.lower()}s.csv"
    return [participant(**row) for row in process_form(path_to_data)]


def _mark_participants_with_no_matches(matrix: List[List[Match]], role_as_str: str):
    for row in matrix:
        if all([match.disallowed for match in row]):
            row[0].__getattribute__(role_as_str).has_no_match = True
            logging.debug(
                f"Participant {row[0].__getattribute__(role_as_str).email} has no matches"
            )


def create_matches(
    mentor_list: List[Mentor],
    mentee_list: List[Mentee],
    weightings: Union[None, Dict[str, int]] = None,
) -> List[List[Match]]:
    def _can_match(participant: Union[Mentor, Mentee]):
        return not participant.has_no_match

    preliminary_matches = generate_match_matrix(mentor_list, mentee_list, weightings)
    _mark_participants_with_no_matches(preliminary_matches, "mentor")
    _mark_participants_with_no_matches(transpose_matrix(preliminary_matches), "mentee")
    return generate_match_matrix(
        list(filter(_can_match, mentor_list)),
        list(filter(_can_match, mentee_list)),
        weightings,
    )


def prepare_matrix(matches: List[List[Match]]) -> Matrix:
    prepared_matrix = make_cost_matrix(
        matches,
        lambda match: sys.maxsize - match.score,
    )
    return prepared_matrix


def transpose_matrix(matrix):
    return [list(row) for row in zip(*matrix)]


def calculate_matches(prepared_matrix: Matrix):
    algorithm = Munkres()
    return algorithm.compute(prepared_matrix)


def match_and_assign_participants(
    participant_lists: Tuple[List[Mentor], List[Mentee]],
    weightings: Union[Dict[str, int], None] = None,
) -> Tuple[List[Mentor], List[Mentee]]:
    matches = create_matches(participant_lists[0], participant_lists[1], weightings)
    for successful_match in calculate_matches(prepare_matrix(matches)):
        match = matches[successful_match[0]][successful_match[1]]
        match.mark_successful()
    return participant_lists[0], participant_lists[1]


def process_data(
    mentors: List[Mentor], mentees: List[Mentee]
) -> Tuple[List[Mentor], List[Mentee]]:
    return reduce(
        match_and_assign_participants,
        [
            None,
            {"profession": 4, "grade": 3, "unmatched bonus": 50},
            {"profession": 0, "grade": 3, "unmatched bonus": 100},
        ],
        (mentors, mentees),
    )


def conduct_matching_from_file(path_to_data) -> Tuple[List[Mentor], List[Mentee]]:
    mentors = create_participant_list_from_path(Mentor, path_to_data)
    mentees = create_participant_list_from_path(Mentee, path_to_data)
    return process_data(mentors, mentees)


def create_mailing_list(
    participant_list: List[Union[Mentor, Mentee]], output_folder: Path
):
    file_name = f"{type(participant_list[0]).__name__.lower()}s-list.csv"
    file = output_folder.joinpath(file_name)
    list_participants_as_dicts = [
        participant.to_dict_for_output() for participant in participant_list
    ]
    field_headings = list_participants_as_dicts[0].keys()
    length_headings = len(field_headings)
    for participant in list_participants_as_dicts:
        if len(participant.keys()) > length_headings:
            field_headings = participant.keys()
    try:
        os.mkdir(output_folder)
    except FileExistsError:
        pass
    with open(file, "w", newline="") as output_file:
        writer = csv.DictWriter(output_file, fieldnames=list(field_headings))
        writer.writeheader()
        for participant in list_participants_as_dicts:
            writer.writerow(participant)
